/*
 * (C) Copyright 2000-2003
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 *  Command Processor Table
 */

#include <common.h>
#include <command.h>

#define FAILED 1
#define PASSED 0
#define CYG_ADDRWORD unsigned long
#define cyg_uint32 unsigned int
#define cyg_uint8  unsigned char
#define cyg_uint16 unsigned short

extern void DurTimeSet(ulong);
extern int onesTest(CYG_ADDRWORD *);
extern int Addr32(cyg_uint32 *, cyg_uint32 *, CYG_ADDRWORD  *);
extern int Bar32(cyg_uint32 *, cyg_uint32 *, CYG_ADDRWORD *);
extern int Addr8(char *, char  *, CYG_ADDRWORD *);
extern int Bar8(char *, char *, CYG_ADDRWORD *);
extern void dumpMem(CYG_ADDRWORD badAddr);
extern int ABar32(cyg_uint32 *,cyg_uint32 *,CYG_ADDRWORD *);	
extern int LoopMemTest(CYG_ADDRWORD , CYG_ADDRWORD ,ulong );
extern void memtst_print_banner(void);
extern void memtst_usage(void);

extern flash_info_t flash_info[];
extern unsigned int flash_in_which_sec (flash_info_t *, unsigned int );
extern  unsigned char* ag7100_mac_addr_loc(void);
extern flash_info_t flash_info[];
extern unsigned long crc32 (unsigned long,const unsigned char *,unsigned int);
extern int do_macset  (cmd_tbl_t *, int , int , char *argv[]);	

int
do_version (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	extern char version_string[];
	printf ("\n%s\n", version_string);
	return 0;
}

int do_wpspinset(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	char wpspin[8] = {0};
	int s_first, s_last;
	
	if(argc != 2){
		printf("%s\n",cmdtp->help);
		return -1;
    }
	strncpy(wpspin, argv[1], 8);
	memcpy((void *)MEM_LOAD_ADDR, (void *)CFG_FLASH_ART_BASE, CFG_FLASH_ART_PARTITION_SIZE);
	memcpy((void *)(MEM_LOAD_ADDR+WPSPIN_OFFSET), wpspin, 8);
	printf("burn wpspin into ART block.\n");
	s_first = flash_in_which_sec (&flash_info[0], CFG_FLASH_ART_BASE - flash_info[0].start[0]);
	s_last = s_first;
	flash_erase(&flash_info[0],s_first,s_last);
	flash_write((uchar *)MEM_LOAD_ADDR, CFG_FLASH_ART_BASE, CFG_FLASH_ART_PARTITION_SIZE);
	printf("Done.\n");
	return 0;
	
}
static void print_banner(void)
{
	 printf("\nUBOOT-1.1.4 MACSET TOOL copyright DNI . \n");
}
static void usage(void)
{
	printf("Usage:\n macset lan(wan)/eth0(eth1) macaddress\n");
	printf("For instance : macset lan(eth0) 03:EE:02:EF:77:87\n");
	printf("The MAC address can not be all 0x0 or all 0xFF\n");
}
int do_macset(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	unsigned char *eeprom,*tmp;
	int i = 0,j=0;
	int val = 0, mac_offset;
	ulong artaddr = 0xbf3f0000;
	ulong sum=0;
	unsigned int s_first,s_end;
	unsigned char mac[6]={0xff,0xff,0xff,0xff,0xff,0xff};
	if(argc != 3){
		print_banner();
		usage();
		return -1;
	}
	tmp = argv[2];
	while(tmp[i]){
		if(tmp[i] == ':')
		{
			mac[j++] = val;
			i++;
			sum += val;
			val = 0;
			continue;
		}
		if((tmp[i] >= '0')&& (tmp[i] <= '9'))
			val = val*16 + (tmp[i] - '0');
		else if((tmp[i] >='a')&&(tmp[i] <= 'f'))
			val = val*16 + (tmp[i] - 'a') + 10;
		else if((tmp[i] >= 'A')&&(tmp[i] <= 'F'))
			val = val*16 + (tmp[i] - 'A') + 10;
		else {
			print_banner();
			usage();
			return -1;
		}
		i ++;
	}
	mac[j]= val;
	sum += val;

	if(j != 5 || sum == 0 || sum == 0xff*6){
		print_banner();
		usage();
		return -1;
	}
	memmove((uchar *)MEM_LOAD_ADDR,(uchar*)artaddr,64*1024);
	s_first = flash_in_which_sec (&flash_info[0], artaddr - flash_info[0].start[0]);
	s_end = s_first;
	
	if(strcmp(argv[1],"lan") == 0 || strcmp(argv[1],"eth0") == 0){
		mac_offset = 0;
	}else if(strcmp(argv[1],"wan") == 0 || strcmp(argv[1],"eth1") == 0){
		mac_offset = 6;
	}else{
		printf("unknown ethernet device\n");
		return -1;
	}

	/* Erase the partition first */
	flash_erase (&flash_info[0], s_first, s_end);
	
	/* Copy MAC address to correct offset */
	eeprom = ag7100_mac_addr_loc();
	eeprom = eeprom - artaddr + MEM_LOAD_ADDR;
	eeprom += mac_offset;
	memcpy(eeprom,mac,6);
	
	/* Write back to flash */
	flash_write ((uchar *)MEM_LOAD_ADDR, artaddr , 64*1024);

	return 0;
}
int
do_testmem (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	ulong dur_time;
	CYG_ADDRWORD startAddr, endAddr,badAddr;  // Addr test failed at

	if(argc < 3){
		memtst_print_banner();
		memtst_usage();
		return FAILED;
	}
	startAddr = simple_strtoul(argv[1], NULL, 16);
	endAddr = simple_strtoul(argv[2], NULL, 16);
	dur_time = simple_strtoul(argv[3],NULL,10);
	if ((startAddr < CFG_MEMTEST_START)||(endAddr > CFG_MEMTEST_END))
	{
		memtst_print_banner();
		memtst_usage();
		return FAILED;
	}
	printf("\nWalking Bit test: ");
	reset_timer();
	if(dur_time <= 21)
		DurTimeSet(dur_time*CFG_HZ);
	if (onesTest((CYG_ADDRWORD  *)startAddr) == FAILED)
		goto failed;
	printf("passed");
	printf("\n32-bit address test: ");
	if (Addr32((cyg_uint32 *)startAddr, (cyg_uint32 *)endAddr, &badAddr) == FAILED)
		goto failed;
	printf("passed");
	printf("\n32-bit address reverse test: ");
	if (Bar32((cyg_uint32 *)startAddr, (cyg_uint32 *)endAddr, &badAddr) == FAILED)
		goto failed;
	printf("passed");
	printf("\nByte address test: ");
	if (Addr8((cyg_uint8 *)startAddr, (cyg_uint8 *)endAddr, &badAddr) == FAILED)
		goto failed;
	printf("passed");
	printf("\nByte address reverse test: ");
	if (Bar8((cyg_uint8 *)startAddr, (cyg_uint8 *)endAddr, &badAddr) == FAILED)
		goto failed;
	printf("passed\n");
	printf("begin to constant pressure testing last for %lu(second)\n",dur_time);
	if(LoopMemTest((CYG_ADDRWORD *)startAddr, (CYG_ADDRWORD *)endAddr,dur_time) == FAILED)
		goto failed;
	return 0;
failed:
	printf("failed");
	dumpMem(badAddr);
	return 1;
}



int
do_checksum (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	unsigned int s_first,s_end;
	ulong checksum;
	ulong addr = CFG_FLASH_ROOTFS_BASE;
	checksum = crc32(0, (unchar *)addr,ROOTFS_LEN);
	
	s_first = flash_in_which_sec (&flash_info[0], CFG_FLASH_ROOTFS_CHECKSUM_BASE - flash_info[0].start[0]);
	s_end = flash_in_which_sec (&flash_info[0], CFG_FLASH_ROOTFS_CHECKSUM_BASE + CFG_FLASH_ROOTFS_CHECKSUM_PARTITION_SIZE - flash_info[0].start[0]) - 1;
	printf ("Erase %d - %d sectors... ", s_first, s_end);
	flash_erase (&flash_info[0], s_first, s_end);
	printf ("Copy rootfs checksum to Flash... ");
	flash_write ((uchar *)(&checksum),CFG_FLASH_ROOTFS_CHECKSUM_BASE,sizeof(checksum));
	printf ("Done\n");
	return 0;
}



U_BOOT_CMD(
	version,	1,		1,	do_version,
 	"version - print monitor version\n",
	NULL
);

U_BOOT_CMD(
	 rfschecksum,		1,		1,		do_checksum,
	"rfschecksum - write checksum to flash\n",
				        NULL
);

U_BOOT_CMD(
		memtest,        CFG_MAXARGS,              1,      do_testmem,
		"memtest - test memory validity \n",
		"Usage: \n memtest [beginaddr] [endaddr] [durtime(sec)]\n"
		"valid memory test address range from 0x80200000 to 0x83800000\n"
		"\tfor instance: memtest 0x80300000 0x83800000 5000\n"
);

U_BOOT_CMD(
		macset,		CFG_MAXARGS,		1,		do_macset,
		"macset - set netinterface macaddress\n",
		"Usage:\n macset wan(lan)/eth1(eth0) macaddress\n"
		"for instance:macset lan(eth0) 03:07:09:EF:EE:EF\n"
);

U_BOOT_CMD(
		wpspinset,CFG_MAXARGS,		1,		do_wpspinset,
		"wpspinset - set wpspin number\n",
		"Usage:\n wpspinset number\n"
		"for instance: wpspinset 12345678\n"
);

#if (CONFIG_COMMANDS & CFG_CMD_ECHO)

int
do_echo (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	int i, putnl = 1;

	for (i = 1; i < argc; i++) {
		char *p = argv[i], c;

		if (i > 1)
			putc(' ');
		while ((c = *p++) != '\0') {
			if (c == '\\' && *p == 'c') {
				putnl = 0;
				p++;
			} else {
				putc(c);
			}
		}
	}

	if (putnl)
		putc('\n');
	return 0;
}

U_BOOT_CMD(
	echo,	CFG_MAXARGS,	1,	do_echo,
 	"echo    - echo args to console\n",
 	"[args..]\n"
	"    - echo args to console; \\c suppresses newline\n"
);

#endif	/*  CFG_CMD_ECHO */

#ifdef CFG_HUSH_PARSER

int
do_test (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	char **ap;
	int left, adv, expr, last_expr, neg, last_cmp;

	/* args? */
	if (argc < 3)
		return 1;

#if 0
	{
		printf("test:");
		left = 1;
		while (argv[left])
			printf(" %s", argv[left++]);
	}
#endif

	last_expr = 0;
	left = argc - 1; ap = argv + 1;
	if (left > 0 && strcmp(ap[0], "!") == 0) {
		neg = 1;
		ap++;
		left--;
	} else
		neg = 0;

	expr = -1;
	last_cmp = -1;
	last_expr = -1;
	while (left > 0) {

		if (strcmp(ap[0], "-o") == 0 || strcmp(ap[0], "-a") == 0)
			adv = 1;
		else if (strcmp(ap[0], "-z") == 0 || strcmp(ap[0], "-n") == 0)
			adv = 2;
		else
			adv = 3;

		if (left < adv) {
			expr = 1;
			break;
		}

		if (adv == 1) {
			if (strcmp(ap[0], "-o") == 0) {
				last_expr = expr;
				last_cmp = 0;
			} else if (strcmp(ap[0], "-a") == 0) {
				last_expr = expr;
				last_cmp = 1;
			} else {
				expr = 1;
				break;
			}
		}

		if (adv == 2) {
			if (strcmp(ap[0], "-z") == 0)
				expr = strlen(ap[1]) == 0 ? 1 : 0;
			else if (strcmp(ap[0], "-n") == 0)
				expr = strlen(ap[1]) == 0 ? 0 : 1;
			else {
				expr = 1;
				break;
			}

			if (last_cmp == 0)
				expr = last_expr || expr;
			else if (last_cmp == 1)
				expr = last_expr && expr;
			last_cmp = -1;
		}

		if (adv == 3) {
			if (strcmp(ap[1], "=") == 0)
				expr = strcmp(ap[0], ap[2]) == 0;
			else if (strcmp(ap[1], "!=") == 0)
				expr = strcmp(ap[0], ap[2]) != 0;
			else if (strcmp(ap[1], ">") == 0)
				expr = strcmp(ap[0], ap[2]) > 0;
			else if (strcmp(ap[1], "<") == 0)
				expr = strcmp(ap[0], ap[2]) < 0;
			else if (strcmp(ap[1], "-eq") == 0)
				expr = simple_strtol(ap[0], NULL, 10) == simple_strtol(ap[2], NULL, 10);
			else if (strcmp(ap[1], "-ne") == 0)
				expr = simple_strtol(ap[0], NULL, 10) != simple_strtol(ap[2], NULL, 10);
			else if (strcmp(ap[1], "-lt") == 0)
				expr = simple_strtol(ap[0], NULL, 10) < simple_strtol(ap[2], NULL, 10);
			else if (strcmp(ap[1], "-le") == 0)
				expr = simple_strtol(ap[0], NULL, 10) <= simple_strtol(ap[2], NULL, 10);
			else if (strcmp(ap[1], "-gt") == 0)
				expr = simple_strtol(ap[0], NULL, 10) > simple_strtol(ap[2], NULL, 10);
			else if (strcmp(ap[1], "-ge") == 0)
				expr = simple_strtol(ap[0], NULL, 10) >= simple_strtol(ap[2], NULL, 10);
			else {
				expr = 1;
				break;
			}

			if (last_cmp == 0)
				expr = last_expr || expr;
			else if (last_cmp == 1)
				expr = last_expr && expr;
			last_cmp = -1;
		}

		ap += adv; left -= adv;
	}

	if (neg)
		expr = !expr;

	expr = !expr;

#if 0
	printf(": returns %d\n", expr);
#endif

	return expr;
}

U_BOOT_CMD(
	test,	CFG_MAXARGS,	1,	do_test,
 	"test    - minimal test like /bin/sh\n",
 	"[args..]\n"
	"    - test functionality\n"
);

int
do_exit (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	int r;

	r = 0;
	if (argc > 1)
		r = simple_strtoul(argv[1], NULL, 10);

	return -r - 2;
}

U_BOOT_CMD(
	exit,	2,	1,	do_exit,
 	"exit    - exit script\n",
	"    - exit functionality\n"
);


#endif

/*
 * Use puts() instead of printf() to avoid printf buffer overflow
 * for long help messages
 */
int do_help (cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{
	int i;
	int rcode = 0;

	if (argc == 1) {	/*show list of commands */

		int cmd_items = &__u_boot_cmd_end -
				&__u_boot_cmd_start;	/* pointer arith! */
		cmd_tbl_t *cmd_array[cmd_items];
		int i, j, swaps;

		/* Make array of commands from .uboot_cmd section */
		cmdtp = &__u_boot_cmd_start;
		for (i = 0; i < cmd_items; i++) {
			cmd_array[i] = cmdtp++;
		}

		/* Sort command list (trivial bubble sort) */
		for (i = cmd_items - 1; i > 0; --i) {
			swaps = 0;
			for (j = 0; j < i; ++j) {
				if (strcmp (cmd_array[j]->name,
					    cmd_array[j + 1]->name) > 0) {
					cmd_tbl_t *tmp;
					tmp = cmd_array[j];
					cmd_array[j] = cmd_array[j + 1];
					cmd_array[j + 1] = tmp;
					++swaps;
				}
			}
			if (!swaps)
				break;
		}

		/* print short help (usage) */
		for (i = 0; i < cmd_items; i++) {
			const char *usage = cmd_array[i]->usage;

			/* allow user abort */
			if (ctrlc ())
				return 1;
			if (usage == NULL)
				continue;
			puts (usage);
		}
		return 0;
	}
	/*
	 * command help (long version)
	 */
	for (i = 1; i < argc; ++i) {
		if ((cmdtp = find_cmd (argv[i])) != NULL) {
#ifdef	CFG_LONGHELP
			/* found - print (long) help info */
			puts (cmdtp->name);
			putc (' ');
			if (cmdtp->help) {
				puts (cmdtp->help);
			} else {
				puts ("- No help available.\n");
				rcode = 1;
			}
			putc ('\n');
#else	/* no long help available */
			if (cmdtp->usage)
				puts (cmdtp->usage);
#endif	/* CFG_LONGHELP */
		} else {
			printf ("Unknown command '%s' - try 'help'"
				" without arguments for list of all"
				" known commands\n\n", argv[i]
					);
			rcode = 1;
		}
	}
	return rcode;
}


U_BOOT_CMD(
	help,	CFG_MAXARGS,	1,	do_help,
 	"help    - print online help\n",
 	"[command ...]\n"
 	"    - show help information (for 'command')\n"
 	"'help' prints online help for the monitor commands.\n\n"
 	"Without arguments, it prints a short usage message for all commands.\n\n"
 	"To get detailed help information for specific commands you can type\n"
  "'help' with one or more command names as arguments.\n"
);

/* This do not ust the U_BOOT_CMD macro as ? can't be used in symbol names */
#ifdef  CFG_LONGHELP
cmd_tbl_t __u_boot_cmd_question_mark Struct_Section = {
	"?",	CFG_MAXARGS,	1,	do_help,
 	"?       - alias for 'help'\n",
	NULL
};
#else
cmd_tbl_t __u_boot_cmd_question_mark Struct_Section = {
	"?",	CFG_MAXARGS,	1,	do_help,
 	"?       - alias for 'help'\n"
};
#endif /* CFG_LONGHELP */

/***************************************************************************
 * find command table entry for a command
 */
cmd_tbl_t *find_cmd (const char *cmd)
{
	cmd_tbl_t *cmdtp;
	cmd_tbl_t *cmdtp_temp = &__u_boot_cmd_start;	/*Init value */
	const char *p;
	int len;
	int n_found = 0;

	/*
	 * Some commands allow length modifiers (like "cp.b");
	 * compare command name only until first dot.
	 */
	len = ((p = strchr(cmd, '.')) == NULL) ? strlen (cmd) : (p - cmd);

	for (cmdtp = &__u_boot_cmd_start;
	     cmdtp != &__u_boot_cmd_end;
	     cmdtp++) {
		if (strncmp (cmd, cmdtp->name, len) == 0) {
			if (len == strlen (cmdtp->name))
				return cmdtp;	/* full match */

			cmdtp_temp = cmdtp;	/* abbreviated command ? */
			n_found++;
		}
	}
	if (n_found == 1) {			/* exactly one match */
		return cmdtp_temp;
	}

	return NULL;	/* not found or ambiguous command */
}

#ifdef CONFIG_AUTO_COMPLETE

int var_complete(int argc, char *argv[], char last_char, int maxv, char *cmdv[])
{
	static char tmp_buf[512];
	int space;

	space = last_char == '\0' || last_char == ' ' || last_char == '\t';

	if (space && argc == 1)
		return env_complete("", maxv, cmdv, sizeof(tmp_buf), tmp_buf);

	if (!space && argc == 2)
		return env_complete(argv[1], maxv, cmdv, sizeof(tmp_buf), tmp_buf);

	return 0;
}

static void install_auto_complete_handler(const char *cmd,
		int (*complete)(int argc, char *argv[], char last_char, int maxv, char *cmdv[]))
{
	cmd_tbl_t *cmdtp;

	cmdtp = find_cmd(cmd);
	if (cmdtp == NULL)
		return;

	cmdtp->complete = complete;
}

void install_auto_complete(void)
{
	install_auto_complete_handler("printenv", var_complete);
	install_auto_complete_handler("setenv", var_complete);
#if (CONFIG_COMMANDS & CFG_CMD_RUN)
	install_auto_complete_handler("run", var_complete);
#endif
}

/*************************************************************************************/

static int complete_cmdv(int argc, char *argv[], char last_char, int maxv, char *cmdv[])
{
	cmd_tbl_t *cmdtp;
	const char *p;
	int len, clen;
	int n_found = 0;
	const char *cmd;

	/* sanity? */
	if (maxv < 2)
		return -2;

	cmdv[0] = NULL;

	if (argc == 0) {
		/* output full list of commands */
		for (cmdtp = &__u_boot_cmd_start; cmdtp != &__u_boot_cmd_end; cmdtp++) {
			if (n_found >= maxv - 2) {
				cmdv[n_found++] = "...";
				break;
			}
			cmdv[n_found++] = cmdtp->name;
		}
		cmdv[n_found] = NULL;
		return n_found;
	}

	/* more than one arg or one but the start of the next */
	if (argc > 1 || (last_char == '\0' || last_char == ' ' || last_char == '\t')) {
		cmdtp = find_cmd(argv[0]);
		if (cmdtp == NULL || cmdtp->complete == NULL) {
			cmdv[0] = NULL;
			return 0;
		}
		return (*cmdtp->complete)(argc, argv, last_char, maxv, cmdv);
	}

	cmd = argv[0];
	/*
	 * Some commands allow length modifiers (like "cp.b");
	 * compare command name only until first dot.
	 */
	p = strchr(cmd, '.');
	if (p == NULL)
		len = strlen(cmd);
	else
		len = p - cmd;

	/* return the partial matches */
	for (cmdtp = &__u_boot_cmd_start; cmdtp != &__u_boot_cmd_end; cmdtp++) {

		clen = strlen(cmdtp->name);
		if (clen < len)
			continue;

		if (memcmp(cmd, cmdtp->name, len) != 0)
			continue;

		/* too many! */
		if (n_found >= maxv - 2) {
			cmdv[n_found++] = "...";
			break;
		}

		cmdv[n_found++] = cmdtp->name;
	}

	cmdv[n_found] = NULL;
	return n_found;
}

static int make_argv(char *s, int argvsz, char *argv[])
{
	int argc = 0;

	/* split into argv */
	while (argc < argvsz - 1) {

		/* skip any white space */
		while ((*s == ' ') || (*s == '\t'))
			++s;

		if (*s == '\0') 	/* end of s, no more args	*/
			break;

		argv[argc++] = s;	/* begin of argument string	*/

		/* find end of string */
		while (*s && (*s != ' ') && (*s != '\t'))
			++s;

		if (*s == '\0')		/* end of s, no more args	*/
			break;

		*s++ = '\0';		/* terminate current arg	 */
	}
	argv[argc] = NULL;

	return argc;
}

static void print_argv(const char *banner, const char *leader, const char *sep, int linemax, char *argv[])
{
	int ll = leader != NULL ? strlen(leader) : 0;
	int sl = sep != NULL ? strlen(sep) : 0;
	int len, i;

	if (banner) {
		puts("\n");
		puts(banner);
	}

	i = linemax;	/* force leader and newline */
	while (*argv != NULL) {
		len = strlen(*argv) + sl;
		if (i + len >= linemax) {
			puts("\n");
			if (leader)
				puts(leader);
			i = ll - sl;
		} else if (sep)
			puts(sep);
		puts(*argv++);
		i += len;
	}
	printf("\n");
}

static int find_common_prefix(char *argv[])
{
	int i, len;
	char *anchor, *s, *t;

	if (*argv == NULL)
		return 0;

	/* begin with max */
	anchor = *argv++;
	len = strlen(anchor);
	while ((t = *argv++) != NULL) {
		s = anchor;
		for (i = 0; i < len; i++, t++, s++) {
			if (*t != *s)
				break;
		}
		len = s - anchor;
	}
	return len;
}

static char tmp_buf[CFG_CBSIZE];	/* copy of console I/O buffer	*/

int cmd_auto_complete(const char *const prompt, char *buf, int *np, int *colp)
{
	int n = *np, col = *colp;
	char *argv[CFG_MAXARGS + 1];		/* NULL terminated	*/
	char *cmdv[20];
	char *s, *t;
	const char *sep;
	int i, j, k, len, seplen, argc;
	int cnt;
	char last_char;

	if (strcmp(prompt, CFG_PROMPT) != 0)
		return 0;	/* not in normal console */

	cnt = strlen(buf);
	if (cnt >= 1)
		last_char = buf[cnt - 1];
	else
		last_char = '\0';

	/* copy to secondary buffer which will be affected */
	strcpy(tmp_buf, buf);

	/* separate into argv */
	argc = make_argv(tmp_buf, sizeof(argv)/sizeof(argv[0]), argv);

	/* do the completion and return the possible completions */
	i = complete_cmdv(argc, argv, last_char, sizeof(cmdv)/sizeof(cmdv[0]), cmdv);

	/* no match; bell and out */
	if (i == 0) {
		if (argc > 1)	/* allow tab for non command */
			return 0;
		putc('\a');
		return 1;
	}

	s = NULL;
	len = 0;
	sep = NULL;
	seplen = 0;
	if (i == 1) { /* one match; perfect */
		k = strlen(argv[argc - 1]);
		s = cmdv[0] + k;
		len = strlen(s);
		sep = " ";
		seplen = 1;
	} else if (i > 1 && (j = find_common_prefix(cmdv)) != 0) {	/* more */
		k = strlen(argv[argc - 1]);
		j -= k;
		if (j > 0) {
			s = cmdv[0] + k;
			len = j;
		}
	}

	if (s != NULL) {
		k = len + seplen;
		/* make sure it fits */
		if (n + k >= CFG_CBSIZE - 2) {
			putc('\a');
			return 1;
		}

		t = buf + cnt;
		for (i = 0; i < len; i++)
			*t++ = *s++;
		if (sep != NULL)
			for (i = 0; i < seplen; i++)
				*t++ = sep[i];
		*t = '\0';
		n += k;
		col += k;
		puts(t - k);
		if (sep == NULL)
			putc('\a');
		*np = n;
		*colp = col;
	} else {
		print_argv(NULL, "  ", " ", 78, cmdv);

		puts(prompt);
		puts(buf);
	}
	return 1;
}

#endif
